/*
    SSSD

    Kerberos 5 Backend Module -- Utilities

    Authors:
        Sumit Bose <sbose@redhat.com>

    Copyright (C) 2009 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "util/util.h"
#include <grp.h>

errno_t become_user(uid_t uid, gid_t gid, bool keep_set_uid)
{
    uid_t cuid;
    int ret = EOK;

    DEBUG(SSSDBG_FUNC_DATA,
          "Trying to become user [%"SPRIuid"][%"SPRIgid"].\n", uid, gid);

    cuid = geteuid();
    if (uid == cuid) {
        DEBUG(SSSDBG_FUNC_DATA, "Already user [%"SPRIuid"].\n", uid);
        goto done;
    }

    sss_set_cap_effective(CAP_SETGID, true);
    sss_set_cap_effective(CAP_SETUID, true);

    /* drop supplementary groups first */
    ret = setgroups(0, NULL);
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_CRIT_FAILURE,
              "setgroups failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }

    ret = setresgid(gid, gid, gid);
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_CRIT_FAILURE,
              "setresgid failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }

    ret = setresuid(uid, uid, (keep_set_uid ? -1 : uid));
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_CRIT_FAILURE,
              "setresuid failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }

done:
    /* clears permitted set as well, so even if suid allows to set
     * euid back to 0 this wouldn't regain any capabilities
     */
    sss_drop_all_caps();

    return ret;
}

struct sss_creds {
    uid_t uid;
    gid_t gid;
    int num_gids;
    gid_t gids[];
};

errno_t restore_creds(struct sss_creds *saved_creds);

/* This is a reversible version of become_user, and returns the saved
 * credentials so that creds can be switched back calling restore_creds */
errno_t switch_creds(TALLOC_CTX *mem_ctx,
                     uid_t uid, gid_t gid,
                     int num_gids, gid_t *gids,
                     struct sss_creds **saved_creds)
{
    struct sss_creds *ssc = NULL;
    int size;
    int ret;
    uid_t myuid;
    uid_t mygid;

    DEBUG(SSSDBG_FUNC_DATA,
          "Trying to switch user to [%"SPRIuid"][%"SPRIgid"].\n", uid, gid);

    myuid = geteuid();
    mygid = getegid();
    if ((myuid == uid) && (mygid == gid)) {
        DEBUG(SSSDBG_FUNC_DATA, "Already requested user, nothing to do.\n");
        return EOK;
    }

    if (saved_creds) { /* save current user credentials */
        size = getgroups(0, NULL);
        if (size == -1) {
            ret = errno;
            DEBUG(SSSDBG_CRIT_FAILURE, "getgroups() failed! (%d, %s)\n",
                                        ret, strerror(ret));
            return ret;
        }

        ssc = talloc_size(mem_ctx,
                          (sizeof(struct sss_creds) + size * sizeof(gid_t)));
        if (!ssc) {
            DEBUG(SSSDBG_CRIT_FAILURE, "Allocation failed!\n");
            return ENOMEM;
        }

        ssc->num_gids = size;
        size = getgroups(ssc->num_gids, ssc->gids);
        if (size != ssc->num_gids) {
            DEBUG(SSSDBG_CRIT_FAILURE, "2nd getgroups() returned different list!");
            talloc_free(ssc);
            return EINVAL;
        }

        ssc->uid = myuid;
        ssc->gid = mygid;
    }

    /* 'man capabilities':
     * - If the effective user ID is changed from 0 to nonzero,
     *   then all capabilities are cleared from the effective set.
     *
     * - If the effective user ID is changed from nonzero to 0,
     *   then the permitted set is copied to the effective set.
     *
     * The above means that behavior depends on if SSSD service user
     * is 'root' or 'sssd'.
     * Manage effective capabilities explicitly to avoid euid change
     * effects.
     */

    sss_set_cap_effective(CAP_SETGID, true); /* if this fails then next op fails anyway */
    ret = setgroups(num_gids, gids);
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_OP_FAILURE,
              "setgroups failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }
    ret = setresgid(-1, gid, -1);
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_OP_FAILURE,
              "setresgid failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }

    sss_set_cap_effective(CAP_SETUID, true);
    ret = setresuid(-1, uid, -1);
    if (ret == -1) {
        ret = errno;
        DEBUG(SSSDBG_OP_FAILURE,
              "setresuid failed [%d][%s].\n", ret, strerror(ret));
        goto done;
    }

    ret = 0;

done:
    sss_set_cap_effective(CAP_SETGID, false);
    sss_set_cap_effective(CAP_SETUID, false); /* might be no-op but don't bother */
    if (ret) {
        /* attempt to restore creds first */
        restore_creds(ssc);
        talloc_free(ssc);
    } else if (saved_creds) {
        *saved_creds = ssc;
    }
    return ret;
}

errno_t restore_creds(struct sss_creds *saved_creds)
{
    if (saved_creds == NULL) {
        /* In case save_creds was saved with the UID already dropped */
        return EOK;
    }

    return switch_creds(saved_creds,
                        saved_creds->uid,
                        saved_creds->gid,
                        saved_creds->num_gids,
                        saved_creds->gids, NULL);
}
